--[[
Copyright 2010, 2011 William Matthews. All rights reserved.

See LICENSE for license terms.
]]

local CB = select(2,...)
local P = CB.Private

local DS = DataStore

function P:PrintGeneric(t,text,indent,max)
	text = text or ""
	indent = indent or ""
	if type(t)=="table" and (not max or max>0) then
		self:Print(indent..text.."{")
		for k,v in pairs(t) do
			self:PrintGeneric(v,k.." = ",indent.."  ",max and max-1)
		end
		self:Print(indent.."}")
	else
		self:Print(indent..text..tostring(t))
	end
end

local function getDebugInfo(default)
	local info = debugstack(3,1,0)
	info = info and strtrim(info)
	return info or default or "DEBUG: "
end

function P:DebugPrint(data,max)
	if not self.db or self.db.global.debug then
		self:PrintGeneric(data,getDebugInfo(),"",max or 10)
	end
end

function P.genericFailCB(...)
	if not P.db or P.db.global.debug then
		local info = debugstack(2,1,0)
		info = info and strtrim(info)
		P:PrintGeneric({...},getDebugInfo(),"",10)
	end
end

-- Takes an item link or an item ID and returns an item ID
function P.idOfItem(item)
	if item==nil then
		return nil
	elseif type(item)=="number" then
		return item
	else
		return tonumber(item:match("item:(%d+)"))	
	end
end

function P.idOfSpell(spell)
	if spell==nil then
		return nil
	elseif type(spell)=="number" then
		return spell
	else
		local link=GetSpellLink(spell)
		return link and tonumber(link:match("spell:(%d+)"))
	end
end

function P.linkOfItem(item)
	local _, link=GetItemInfo(item)
	return link
end

function P.nameOfItem(item)
	return GetItemInfo(P.idOfItem(item)) or "Unknown"
end

function P.coloredNameOfItem(item)
	name, _, quality = GetItemInfo(P.idOfItem(item))
	if quality then
		return format("|c%s%s|r",select(4,GetItemQualityColor(quality)),name)
	else
		return name or "Unknown"
	end
end

function P.moneyString(money)
	local gold = floor(money / 10000)
	local silver = floor((money - gold * 10000) / 100)
	local copper = mod(money, 100)
	local rv = nil
	if gold > 0 then
		rv=format(GOLD_AMOUNT_TEXTURE, gold, 0, 0)
	end
	if silver > 0 then
		if rv then
			rv=rv.." "..format(SILVER_AMOUNT_TEXTURE, silver, 0, 0)
		else
			rv=format(SILVER_AMOUNT_TEXTURE, silver, 0, 0)
		end
	end
	if copper > 0 then
		if rv then
			rv=rv.." "..format(COPPER_AMOUNT_TEXTURE, copper, 0, 0)
		else
			rv=format(COPPER_AMOUNT_TEXTURE, copper, 0, 0)
		end
	end
	if rv then
		return rv
	else
		return format(COPPER_AMOUNT_TEXTURE, 0, 0, 0)
	end
end

function P.moneyOrUnknown(money)
	if money then
		return P.moneyString(money)
	else
		return "unknown"
	end
end

function P:GetTotalCBItemCount(item)
	local id=P.idOfItem(item)
--	local _,link=GetItemInfo(id)
	local owned = 0
	for characterName, character in pairs(DS:GetCharacters()) do
		if self.db.factionrealm.characters[character] then
			local bagCount, bankCount = DS:GetContainerItemCount(character,id)
			bagCount = bagCount or 0
			bankCount = bankCount or 0
			local mailCount = 0
			local numMails = DS:GetNumMails(character) or 0
			for mailId = 1, numMails do
				local sender = DS:GetMailSender(character,mailId)
				local _, mailItemCount, mailItemLink = DS:GetMailInfo(character,mailId)
				sender = sender and DS:GetCharacters()[sender]
				if sender and mailItemLink and self.db.factionrealm.characters[sender] and P.idOfItem(mailItemLink)==id then
					mailCount = mailCount + (mailItemCount or 1)
				end
			end
			local ahCount = DS:GetAuctionHouseItemCount(character,id) or 0
--			DebugPrint(format("%s: bagCount=%s bankCount=%s mailCount=%s ahCount=%s",characterName,bagCount or "NIL",bankCount or "NIL",mailCount or "NIL",ahCount or "NIL"))
			owned = owned + bagCount + bankCount + mailCount + ahCount
		end
	end
	for guildName, guild in pairs(DS:GetGuilds()) do
		if self.db.factionrealm.guilds[guild] then
			local gbCount = DS:GetGuildBankItemCount(guild,id) or 0
--			DebugPrint(format("%s: gbCount=%s",guildName,gbCount or "NIL"))
			owned = owned + gbCount
		end
	end
	return owned
end

-- Queue ADT
local Queue={}
local QueueMT = {__index=Queue}
function Queue.New() 
	local rv={}
	rv.__head=1
	rv.__tail=1
	setmetatable(rv,QueueMT)
	return rv
end

function Queue:Head()
	return self[self.__head]
end

function Queue:Push(x)
	self[self.__tail]=x
	self.__tail=self.__tail+1
end

function Queue:Pop()
	local rv=self[self.__head]
	self[self.__head]=nil -- GC, baby!
	if self.__head<self.__tail then
		self.__head=self.__head+1
	end
	return rv
end

function Queue:Empty()
	return self.__head==self.__tail
end
P.Queue=Queue

function P.setdefault(table,default)
	if type(default)=="function" then
		setmetatable(table,{__index=function(t,k)
			local rv=default(t,k)
			rawset(t,k,rv)
			return rv
		end})
	else
		setmetatable(table,{__index=function(t,k) 
			rawset(t,k,default)
			return default
		end})
	end
end


local Timer = {}
local TimerMT = {__index=Timer}
local function doTimer(t)
	t.__timer=nil
	t.__f(t.__arg)
end
function Timer.new(f,time,arg)
	local rv={}
	rv.__f=f
	rv.__time=math.abs(time)
	rv.__arg=arg
	if time>=0 then
		rv.__timer=P:ScheduleTimer(doTimer,time,rv)
	else
		rv.__timer=nil
	end
	setmetatable(rv,TimerMT)
	return rv
end
function Timer.updateOrNew(timer,f,time,arg)
	if timer then
		timer:Update(time)
		return timer
	else
		return Timer.new(f,time,arg)
	end
end
function Timer:Update(time)
	if self.__timer then
		P:CancelTimer(self.__timer)
	end
	self.__time=time or self.__time
	self.__timer=P:ScheduleTimer(doTimer,self.__time,self)
end
function Timer:Cancel()
	if self.__timer then
		P:CancelTimer(self.__timer)
		self.__timer=nil
	end
end
P.Timer = Timer

local ItemSet= {}
local ItemSetMT = {
	__index = function(t,k)
		local rv = ItemSet[k]
		if rv then return rv end
		assert(type(k)=="number","ItemSet keys must be item IDs.")
		return 0
	end,
	__newindex = function(t,k,v)
		assert(type(k)=="number","ItemSet keys must be item IDs.")
		rawset(t,k,v)
		return v
	end,
}
function ItemSet.new()
	local rv={}
	setmetatable(rv,ItemSetMT)
	return rv
end
function ItemSet:Inc(id,amount)
	self[id] = self[id] + (amount or 1)
end
P.ItemSet = ItemSet

function P.buildRetryFunction(func)
	return function(self,successCB,failCB,...)
		local tries=self.RETRIES
		local args={...}
		local timer,try,process
		process = function(x,...)
			if x then
				return successCB(x,...)
			else
				tries = tries-1
				if tries>0 then
					timer=Timer.updateOrNew(timer,try,self.RETRY_TIME)
				else
					return failCB()
				end
			end
		end
		try = function() return process(func(unpack(args))) end
		return try()
	end
end

P.GetItemInfo=P.buildRetryFunction(GetItemInfo)
P.GetItemSpell=P.buildRetryFunction(GetItemSpell)
P.GetSpellInfo=P.buildRetryFunction(GetSpellInfo)